﻿using EonaCat.HID.EventArguments;
using EonaCat.HID.Managers;
using EonaCat.HID.Managers.Linux;
using EonaCat.HID.Models;
using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.HID
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/license for full license details.

    internal sealed class HidLinux : IHid
    {
        private readonly string _devicePath;
        private FileStream _stream;
        private bool _isOpen;
        private readonly object _lock = new object();
        private CancellationTokenSource _cts;

        public string DevicePath => _devicePath;
        public ushort VendorId { get; private set; }
        public ushort ProductId { get; private set; }
        public string SerialNumber { get; private set; }
        public string Manufacturer { get; private set; }
        public string ProductName { get; private set; }
        public int InputReportByteLength { get; private set; }
        public int OutputReportByteLength { get; private set; }
        public int FeatureReportByteLength { get; private set; }

        public bool IsReadingSupport { get; private set; }
        public bool IsWritingSupport { get; private set; }

        public bool IsConnected => _isOpen;
        public IDictionary<string, object> Capabilities { get; private set; } = new Dictionary<string, object>();

        public event EventHandler<HidDataReceivedEventArgs> OnDataReceived;
        public event EventHandler<HidErrorEventArgs> OnError;

        public HidLinux(string devicePath)
        {
            _devicePath = devicePath ?? throw new ArgumentNullException(nameof(devicePath));
        }

        internal void Setup()
        {
            Open();
            LoadDeviceInfo();
            LoadReportLengths();
            Close();
        }

        public void Open()
        {
            lock (_lock)
            {
                if (_stream != null || _isOpen)
                    return;

                // Open HID device in non-blocking read/write mode
                int fd = NativeMethods.open(_devicePath, NativeMethods.O_RDWR | NativeMethods.O_NONBLOCK);
                if (fd < 0)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this,
                        new IOException($"Failed to open device: {_devicePath}")));
                    return;
                }

                var safeHandle = new Microsoft.Win32.SafeHandles.SafeFileHandle(new IntPtr(fd), ownsHandle: true);

                try
                {
                    _stream = new FileStream(safeHandle, FileAccess.ReadWrite, bufferSize: 64, isAsync: false);
                    _isOpen = true;

                    IsReadingSupport = true;
                    IsWritingSupport = true;
                }
                catch (Exception ex)
                {
                    safeHandle.Dispose();
                    OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                }
            }
        }

        public void Close()
        {
            lock (_lock)
            {
                StopListening();

                if (_stream != null)
                {
                    _stream.Dispose();
                    _stream = null;
                }
            }
        }

        public void Dispose()
        {
            Close();
        }

        private void LoadDeviceInfo()
        {
            string sysPath = GetSysfsDevicePath(_devicePath);
            if (string.IsNullOrEmpty(sysPath))
            {
                return;
            }

            VendorId = ReadHexFileAsUShort(Path.Combine(sysPath, "idVendor"));
            ProductId = ReadHexFileAsUShort(Path.Combine(sysPath, "idProduct"));
            SerialNumber = ReadTextFile(Path.Combine(sysPath, "serial"));
            Manufacturer = ReadTextFile(Path.Combine(sysPath, "manufacturer"));
            ProductName = ReadTextFile(Path.Combine(sysPath, "product"));
        }

        private void LoadReportLengths()
        {
            // As Linux/hidraw interface does not provide direct API for report lengths,
            // We fallback to standard max USB HID report size or fixed size
            // TODO: Maybe we should use hidapi or libusb bindings?

            InputReportByteLength = 64;    // usually max 64 on USB HID
            OutputReportByteLength = 64;
            FeatureReportByteLength = 64;
        }

        private string GetSysfsDevicePath(string devPath)
        {
            // Mapping /dev/hidrawX to sysfs device path: /sys/class/hidraw/hidrawX/device/ 
            var devName = Path.GetFileName(devPath);
            var sysDevicePath = Path.Combine("/sys/class/hidraw", devName, "device");
            if (Directory.Exists(sysDevicePath))
            {
                return sysDevicePath;
            }

            return null;
        }

        private ushort ReadHexFileAsUShort(string path)
        {
            try
            {
                if (File.Exists(path))
                {
                    string text = File.ReadAllText(path).Trim().ToLowerInvariant();
                    if (text.StartsWith("0x"))
                    {
                        text = text.Substring(2);
                    }

                    if (ushort.TryParse(text, System.Globalization.NumberStyles.HexNumber, null, out ushort val))
                    {
                        return val;
                    }
                }
            }
            catch { }
            return 0;
        }

        private string ReadTextFile(string path)
        {
            try
            {
                if (File.Exists(path))
                {
                    return File.ReadAllText(path).Trim();
                }
            }
            catch { }
            return null;
        }

        public async Task WriteOutputReportAsync(HidReport report)
        {
            if (report == null)
            {
                var ex = new ArgumentNullException(nameof(report));
                OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                return;
            }

            if (_stream == null)
            {
                var ex = new InvalidOperationException("Device not open");
                OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                return;
            }

            var data = report.Data;

            await Task.Run(() =>
            {
                lock (_lock)
                {
                    try
                    {
                        _stream.Write(data, 0, data.Length);
                        _stream.Flush();
                    }
                    catch (Exception ex)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                        throw;
                    }
                }
            });
        }

        public async Task<HidReport> ReadInputReportAsync()
        {
            if (_stream == null)
            {
                var ex = new InvalidOperationException("Device not open");
                OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                return null;
            }

            return await Task.Run(() =>
            {
                byte[] buffer = new byte[InputReportByteLength];
                int bytesRead = 0;

                lock (_lock)
                {
                    try
                    {
                        bytesRead = _stream.Read(buffer, 0, buffer.Length);
                    }
                    catch (Exception ex)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                        return null;
                    }
                }

                if (bytesRead <= 0)
                {
                    return null;
                }

                // First byte is report ID, rest is data
                byte reportId = buffer[0];
                byte[] data = new byte[bytesRead - 1];
                Array.Copy(buffer, 1, data, 0, data.Length);

                return new HidReport(reportId, data);
            });
        }


        public async Task SendFeatureReportAsync(HidReport report)
        {
            if (report == null)
                throw new ArgumentNullException(nameof(report));

            // Prepare full report buffer: [reportId][reportData...]
            int size = 1 + (report.Data?.Length ?? 0);
            byte[] buffer = new byte[size];
            buffer[0] = report.ReportId;
            if (report.Data != null && report.Data.Length > 0)
            {
                Array.Copy(report.Data, 0, buffer, 1, report.Data.Length);
            }

            await Task.Run(() =>
            {
                int fd = NativeMethods.open(_devicePath, NativeMethods.O_RDWR);
                if (fd < 0)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException($"Failed to open device: {_devicePath}")));
                    return;
                }

                IntPtr unmanagedBuffer = IntPtr.Zero;
                try
                {
                    unmanagedBuffer = Marshal.AllocHGlobal(size);
                    Marshal.Copy(buffer, 0, unmanagedBuffer, size);

                    int request = NativeMethods._IOC(NativeMethods._IOC_WRITE, 'H', 0x06, size); // HIDIOCSFEATURE
                    int result = NativeMethods.ioctl(fd, request, unmanagedBuffer);
                    if (result < 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException("ioctl HIDIOCSFEATURE failed")));
                    }
                }
                finally
                {
                    if (unmanagedBuffer != IntPtr.Zero)
                        Marshal.FreeHGlobal(unmanagedBuffer);
                    NativeMethods.close(fd);
                }
            });
        }




        public async Task<HidReport> GetFeatureReportAsync(byte reportId)
        {
            const int maxFeatureSize = 256;
            byte[] buffer = new byte[maxFeatureSize];
            buffer[0] = reportId;

            return await Task.Run(() =>
            {
                int fd = NativeMethods.open(_devicePath, NativeMethods.O_RDWR);
                if (fd < 0)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException($"Failed to open device: {_devicePath}")));
                    return new HidReport(0, Array.Empty<byte>());
                }

                IntPtr bufPtr = IntPtr.Zero;
                try
                {
                    bufPtr = Marshal.AllocHGlobal(buffer.Length);
                    Marshal.Copy(buffer, 0, bufPtr, buffer.Length);

                    int request = NativeMethods._IOC(NativeMethods._IOC_READ, 'H', 0x07, buffer.Length); // HIDIOCGFEATURE
                    int result = NativeMethods.ioctl(fd, request, bufPtr);
                    if (result < 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException("ioctl HIDIOCGFEATURE failed")));
                        return new HidReport(0, Array.Empty<byte>());
                    }

                    byte[] actualBuffer = new byte[result];
                    Marshal.Copy(bufPtr, actualBuffer, 0, result);

                    byte actualReportId = actualBuffer.Length > 0 ? actualBuffer[0] : (byte)0;
                    byte[] reportData = actualBuffer.Length > 1 ? new byte[actualBuffer.Length - 1] : Array.Empty<byte>();
                    if (reportData.Length > 0)
                        Array.Copy(actualBuffer, 1, reportData, 0, reportData.Length);

                    return new HidReport(actualReportId, reportData);
                }
                finally
                {
                    if (bufPtr != IntPtr.Zero)
                        Marshal.FreeHGlobal(bufPtr);
                    NativeMethods.close(fd);
                }
            });
        }

        public async Task StartListeningAsync(CancellationToken cancellationToken)
        {
            if (_stream == null)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Device is not open.")));
                return;
            }

            lock (_lock)
            {
                if (_cts != null)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Already listening on this device.")));
                    return;
                }

                _cts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            }

            try
            {
                var token = _cts.Token;

                while (!token.IsCancellationRequested)
                {
                    byte[] buffer = new byte[InputReportByteLength];
                    int bytesRead;

                    try
                    {
                        bytesRead = await _stream.ReadAsync(buffer, 0, buffer.Length, token);
                    }
                    catch (OperationCanceledException)
                    {
                        break;
                    }
                    catch (NotSupportedException)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Reading input reports is not supported on this device.")));
                        break;
                    }
                    catch (Exception ex)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                        break;
                    }

                    if (bytesRead > 0)
                    {
                        var data = new byte[bytesRead];
                        Array.Copy(buffer, data, bytesRead);

                        // Extract reportId and report data
                        byte reportId = data.Length > 0 ? data[0] : (byte)0;

                        byte[] reportData;
                        if (data.Length > 1)
                        {
                            reportData = new byte[data.Length - 1];
                            Array.Copy(data, 1, reportData, 0, data.Length - 1);
                        }
                        else
                        {
                            reportData = Array.Empty<byte>();
                        }

                        var hidReport = new HidReport(reportId, reportData);
                        OnDataReceived?.Invoke(this, new HidDataReceivedEventArgs(this, hidReport));
                    }
                    else
                    {
                        try
                        {
                            await Task.Delay(5, token);
                        }
                        catch (OperationCanceledException)
                        {
                            break;
                        }
                    }
                }
            }
            finally
            {
                lock (_lock)
                {
                    _cts?.Dispose();
                    _cts = null;
                }
            }
        }

        private void StopListening()
        {
            lock (_lock)
            {
                if (_cts != null)
                {
                    _cts.Cancel();
                    _cts.Dispose();
                    _cts = null;
                }
            }
        }
    }
}